
module FastTransforms

using ToeplitzMatrices, HierarchicalMatrices, LowRankApprox, ProgressMeter,
        AbstractFFTs, SpecialFunctions, FastGaussQuadrature

using FFTW, LinearAlgebra, DSP
import FFTW: r2rFFTWPlan, unsafe_execute!, fftwSingle, fftwDouble, fftwNumber
import FFTW: libfftw3, libfftw3f, PlanPtr, r2rFFTWPlan, plan_r2r!,
                REDFT00, REDFT01, REDFT10, REDFT11,
                RODFT00, RODFT01, RODFT10, RODFT11

const libfftw = libfftw3
const libfftwf = libfftw3f
import LinearAlgebra: Factorization
flipdim(A,d) = reverse(A; dims=d)

import Base: *, \, inv, size, view
import Base: getindex, setindex!, length, axes
import LinearAlgebra: BlasFloat, BlasInt, transpose, adjoint, lmul!, rmul!
import HierarchicalMatrices: HierarchicalMatrix, unsafe_broadcasttimes!
import HierarchicalMatrices: mul!
import HierarchicalMatrices: ThreadSafeVector, threadsafezeros
import LowRankApprox: ColPerm, RowPerm
import AbstractFFTs: Plan

import FastGaussQuadrature: unweightedgausshermite

export cjt, icjt, jjt, plan_cjt, plan_icjt
export leg2cheb, cheb2leg, leg2chebu, ultra2ultra, jac2jac, plan_jac2jac
export normleg2cheb, cheb2normleg, normleg12cheb2, cheb22normleg1
export plan_leg2cheb, plan_cheb2leg
export plan_normleg2cheb, plan_cheb2normleg
export plan_normleg12cheb2, plan_cheb22normleg1

export gaunt

export nufft, nufft1, nufft2, nufft3, inufft1, inufft2
export plan_nufft, plan_nufft1, plan_nufft2, plan_nufft3, plan_inufft1, plan_inufft2

export paduatransform, ipaduatransform, paduatransform!, ipaduatransform!, paduapoints
export plan_paduatransform!, plan_ipaduatransform!

export SlowSphericalHarmonicPlan, FastSphericalHarmonicPlan, ThinSphericalHarmonicPlan
export sph2fourier, fourier2sph, plan_sph2fourier
export sphones, sphzeros, sphrand, sphrandn, sphevaluate

export SlowTriangularHarmonicPlan
export tri2cheb, cheb2tri, plan_tri2cheb
export triones, trizeros, trirand, trirandn, trievaluate

export hermitepoints, weightedhermitetransform, iweightedhermitetransform

# Other module methods and constants:
#export ChebyshevJacobiPlan, jac2cheb, cheb2jac
#export sqrtpi, pochhammer, stirlingseries, stirlingremainder, Aratio, Cratio, Anαβ
#export Cnmαβ, Cnαβ, Cnmλ, Cnλ, Λ, absf, findmindices!
#export plan_clenshawcurtis, clenshawcurtisnodes, clenshawcurtisweights
#export plan_fejer1, fejernodes1, fejerweights1
#export plan_fejer2, fejernodes2, fejerweights2
#export RecurrencePlan, forward_recurrence!, backward_recurrence

lgamma(x) = logabsgamma(x)[1]

include("stepthreading.jl")
include("fftBigFloat.jl")
include("specialfunctions.jl")
include("chebyshevtransform.jl")
include("clenshawcurtis.jl")
include("fejer.jl")
include("recurrence.jl")
include("PaduaTransform.jl")

abstract type FastTransformPlan{D,T} end

include("ChebyshevJacobiPlan.jl")
include("jac2cheb.jl")
include("cheb2jac.jl")

include("ChebyshevUltrasphericalPlan.jl")
include("ultra2cheb.jl")
include("cheb2ultra.jl")
include("nufft.jl")
include("inufft.jl")

include("cjt.jl")

include("toeplitzhankel.jl")
include("hermite.jl")

#leg2cheb(x...)=th_leg2cheb(x...)
#cheb2leg(x...)=th_cheb2leg(x...)
leg2chebu(x...) = th_leg2chebu(x...)
ultra2ultra(x...) = th_ultra2ultra(x...)
jac2jac(x...) = jjt(x...)
plan_jac2jac(x...) = plan_jjt(x...)
plan_ultra2ultra(x...) = th_ultra2ultraplan(x...)

include("hierarchical.jl")
include("SphericalHarmonics/SphericalHarmonics.jl")
include("TriangularHarmonics/TriangularHarmonics.jl")

include("gaunt.jl")


include("precompile.jl")
_precompile_()


##
# FastTransform v0.7 compatibility
##

export plan_cheb2jac, plan_jac2cheb, cheb2jac, jac2cheb

plan_cheb2jac(v...) = plan_icjt(v...)
plan_jac2cheb(v...) = plan_cjt(v...)
cheb2jac(v...) = icjt(v...)
jac2cheb(v...) = cjt(v...)

plan_cheb2leg(::Type{T}, n::Int) where T = th_cheb2legplan(T, n)
plan_leg2cheb(::Type{T}, n::Int) where T = th_leg2chebplan(T, n)


end # module
